from abc import ABC, abstractmethod
from typing import Callable, Mapping, Optional, Sequence, Tuple

import xarray as xr
from datacube.model import Dataset
from datacube.utils.geometry import GeoBox
from odc.algo import to_rgba
from odc.algo.io import load_with_native_transform


class StatsPluginInterface(ABC):
    NAME = "*unset*"
    SHORT_NAME = ""
    VERSION = "0.0.0"
    PRODUCT_FAMILY = "statistics"

    def __init__(self,
                 resampling: str = "bilinear",
                 input_bands: Sequence[str] = [],
                 chunks: Mapping[str, int] = {"y": -1, "x": -1},
                 basis: Optional[str] = None,
                 group_by: str = "solar_day",
                 rgb_bands: Optional[Sequence[str]] = None,
                 rgb_clamp: Tuple[float, float] = (1.0, 3_000.0)
                ):
        self.resampling = resampling
        self.input_bands = input_bands
        self.chunks = chunks
        self.basis = basis
        self.group_by = group_by
        self.rgb_bands = rgb_bands
        self.rgb_clamp = rgb_clamp

    @property
    @abstractmethod
    def measurements(self) -> Tuple[str, ...]:
        pass

    def native_transform(self, xx: xr.Dataset) -> xr.Dataset:
        return xx

    def fuser(self, xx: xr.Dataset) -> xr.Dataset:
        return xx

    def input_data(self, datasets: Sequence[Dataset], geobox: GeoBox) -> xr.Dataset:
        xx = load_with_native_transform(
            datasets,
            bands=self.input_bands,
            geobox=geobox,
            native_transform=self.native_transform,
            basis=self.basis,
            groupby=self.group_by,
            fuser=self.fuser,
            resampling=self.resampling,
            chunks=self.chunks,
        )
        return xx

    @abstractmethod
    def reduce(self, xx: xr.Dataset) -> xr.Dataset:
        pass

    def rgba(self, xx: xr.Dataset) -> Optional[xr.DataArray]:
        """
        Given result of ``.reduce(..)`` optionally produce RGBA preview image
        """
        if self.rgb_bands is None:
            return None
        return to_rgba(xx, clamp=self.rgb_clamp, bands=self.reg_bands)



"""
Fractional Cover Percentiles
"""
from functools import partial
from typing import Optional, Sequence, Tuple, Dict
import xarray as xr
import numpy as np
from odc.algo import keep_good_only
from odc.algo._percentile import xr_quantile_bands
from odc.algo._masking import _xr_fuse, _fuse_mean_np, enum_to_bool
from ._registry import StatsPluginInterface, register

NODATA = -9999 # output NODATA


class StatsTCWPC(StatsPluginInterface):

    NAME = "ga_tcw_percentiles"
    SHORT_NAME = NAME
    VERSION = "0.0.1"
    PRODUCT_FAMILY = "twc_percentiles"

    def __init__(
        self,
        coefficients: Dict[str, float] = {
            'blue': 0.0315, 'green': 0.2021, 'red': 0.3102, 'nir': 0.1594, 'swir1': -0.6806, 'swir2': -0.6109
            },
        input_bands: Sequence[str] = ["blue", "green", "red", "nir", "swir1", "swir2", "fmask", "nbart_contiguity"],
        **kwargs
    ):
        super().__init__(input_bands=input_bands, **kwargs)
        self.coefficients = coefficients

    @property
    def measurements(self) -> Tuple[str, ...]:
        _measurments = ["wet_pc_10", "wet_pc_50", "wet_pc_90"]
        return _measurments

    def native_transform(self, xx):
        """
        Loads data in its native projection.
        """
        bad = enum_to_bool(xx["fmask"], ("nodata", "cloud", "shadow")) # a pixel is bad if any of the cloud, shadow, or no-data value
        bad |= xx["nbart_contiguity"] == 0 # or the nbart contiguity bit is 0
        xx = xx.drop_vars(["fmask", "nbart_contiguity"])
        
        for band in xx.data_vars.keys():
            bad = bad | (xx[band] == -999)

        tcw = sum(coeff * xx[band] for band, coeff in self.coefficients.items())
        tcw.attrs = xx.blue.attrs
        tcw.attrs['nodata'] = NODATA
        
        xx = xx.drop_vars(xx.data_vars.keys())
        xx['wet'] = tcw.astype(np.int16)
        xx = keep_good_only(xx, ~bad, nodata=NODATA)
        return xx

    @staticmethod
    def fuser(xx):
        xx = _xr_fuse(xx, partial(_fuse_mean_np, nodata=NODATA), '')

        return xx
    
    @staticmethod
    def reduce(xx: xr.Dataset) -> xr.Dataset:
        yy = xr_quantile_bands(xx, [0.1, 0.5, 0.9], nodata=NODATA)
        return yy


register("tcw-percentiles", StatsTCWPC)



import pydoc
from typing import Callable, Dict, Type
from functools import partial

from ._base import StatsPluginInterface


PluginFactory = Callable[..., StatsPluginInterface]

_plugins: Dict[str, PluginFactory] = {}


def _new(plugin_class, *args, **kwargs) -> StatsPluginInterface:
    return plugin_class(*args, **kwargs)


def resolve(name: str) -> PluginFactory:
    maker = _plugins.get(name)
    if maker is None:
        maker = pydoc.locate(name)
        if maker is not None:
            if not issubclass(maker, (StatsPluginInterface,)):
                raise ValueError(f"Custom StatsPlugin '{name}' is not derived from StatsPluginInterface")
            return partial(_new, maker)

    if maker is None:
        raise ValueError(f"Failed to resolved named plugin: '{name}'")
    return maker


def register(name: str, plugin_class: Type[StatsPluginInterface]):
    _plugins[name] = partial(_new, plugin_class)


def import_all():
    import importlib

    # TODO: make that more automatic
    modules = [
        "odc.stats.plugins.fc_percentiles",
        "odc.stats.plugins.tcw_percentiles",
        "odc.stats.plugins.gm",
        "odc.stats.plugins.gm_ls_bitmask",
        "odc.stats.plugins.pq",
        "odc.stats.plugins.pq_bitmask",
        "odc.stats.plugins.wofs",
    ]

    for mod in modules:
        try:
            importlib.import_module(mod)
        except ModuleNotFoundError:
            pass




import math
from abc import ABC, abstractmethod
from copy import deepcopy
from dataclasses import dataclass, field
from datetime import datetime, timedelta, timezone
from typing import Any, Dict, List, Optional, Tuple, Union
from uuid import UUID
from pathlib import Path

import pandas as pd
import pystac
import xarray as xr
from datacube.model import Dataset
from datacube.utils.dates import normalise_dt
from datacube.utils.geometry import GeoBox
from odc.index import odc_uuid
from ._text import split_and_check
from pystac.extensions.projection import ProjectionExtension
from toolz import dicttoolz
from rasterio.crs import CRS
import warnings

from eodatasets3.assemble import DatasetAssembler, serialise
from eodatasets3.images import GridSpec

from .plugins import StatsPluginInterface

TileIdx_xy = Tuple[int, int]
TileIdx_txy = Tuple[str, int, int]
TileIdx = Union[TileIdx_txy, TileIdx_xy]

default_href_prefix = "https://collections.dea.ga.gov.au/product"
EXT_TIFF = "tif"  # because "consistency"


def format_datetime(dt: datetime, with_tz=True, timespec="microseconds") -> str:
    dt = normalise_dt(dt)
    dt = dt.isoformat(timespec=timespec)
    if with_tz:
        dt = dt + "Z"
    return dt


@dataclass
class DateTimeRange:

    __slots__ = ("start", "end", "freq")

    def __init__(self, start: Union[str, datetime], freq: Optional[str] = None):
        """

        DateTimeRange('2019-03--P3M')
        DateTimeRange('2019-03', '3M')
        DateTimeRange(datetime(2019, 3, 1), '3M')

        """

        if freq is None:
            assert isinstance(start, str)
            start, freq = split_and_check(start, "--P", 2)
        freq = freq.upper().lstrip("P")
        # Pandas period snaps to frequency resolution, we need to undo that by re-adding the snapping delta
        t0 = pd.Timestamp(start)
        period = pd.Period(t0, freq=freq)
        dt = t0 - period.start_time

        self.freq: str = freq

        self.start: datetime = normalise_dt(t0.to_pydatetime(warn=False))
        self.end: datetime = normalise_dt(
            (period.end_time + dt).to_pydatetime(warn=False)
        )

    @staticmethod
    def year(year: int) -> "DateTimeRange":
        """
        Construct ``DateTimeRange`` covering one whole year.
        """
        return DateTimeRange(datetime(year, 1, 1), "1Y")

    def __str__(self):
        return self.short

    def __repr__(self):
        return f"DateTimeRange({repr(self.start)}, {repr(self.freq)})"

    def dc_query(
        self, pad: Optional[Union[timedelta, float, int]] = None
    ) -> Dict[str, Any]:
        """
        Transform to form understood by datacube

        :param pad: optionally pad the region by X days, or timedelta

        Example: ``dc.load(..., **p.dc_query(pad=0.5))``
        """
        if pad is None:
            return {"time": (self.start, self.end)}

        if isinstance(pad, (int, float)):
            pad = timedelta(days=pad)

        return {"time": (self.start - pad, self.end + pad)}

    @property
    def short(self) -> str:
        """
        Short string representation of the time period.

        Examples: 2019--P1Y, 2020-01--P3M, 2013-03-21--P10D
        """
        freq = self.freq
        dt = self.start
        if freq.endswith("Y") and dt.month == 1 and dt.day == 1:
            return f"{dt.year}--P{freq}"
        elif freq.endswith("M") and dt.day == 1:
            return f"{dt.year}-{dt.month:02d}--P{freq}"
        else:
            return f"{dt.year}-{dt.month:02d}-{dt.day:02d}--P{freq}"

    def __contains__(self, t: datetime) -> bool:
        return self.start <= t <= self.end

    def __lt__(self, t: datetime) -> bool:
        return self.start < t

    def __gt__(self, t: datetime) -> bool:
        return self.end > t

    def to_pandas(self) -> pd.Period:
        """
        Convert to pandas Period object
        """
        return pd.Period(self.start, self.freq)

    def __add__(self, v: int) -> "DateTimeRange":
        p = self.to_pandas() + v
        return DateTimeRange(p.start_time.to_pydatetime(warn=False), self.freq)

    def __sub__(self, v: int) -> "DateTimeRange":
        p = self.to_pandas() - v
        return DateTimeRange(p.start_time.to_pydatetime(warn=False), self.freq)


@dataclass
class OutputProduct:
    name: str
    version: str
    short_name: str
    location: str
    properties: Dict[str, str]
    measurements: Tuple[str, ...]
    href: str = ""
    region_code_format: str = "x{x:02d}y{y:02d}"
    cfg: Any = None
    naming_conventions_values: str = "dea_c3"
    explorer_path: str = "https://explorer.dea.ga.gov.au/"
    inherit_skip_properties: Optional[List[str]] = None
    preview_image_ows_style: Optional[Dict[str, Any]] = None
    classifier: str = "level3"
    maturity: str = "final"
    collection_number: int = 3
    nodata : Optional[Dict[str, int]] = None


    def __post_init__(self):
        if self.href == "":
            self.href = f"{default_href_prefix}/{self.name}"

    def region_code(self, tidx: TileIdx_xy) -> str:
        """
        Render tile index into a string.
        """
        x, y = tidx
        return self.region_code_format.format(x=x, y=y)

    @staticmethod
    def dummy(
        measurements: Tuple[str, ...] = ("red", "green", "blue")
    ) -> "OutputProduct":
        version = "0.0.0"
        name = "dummy"
        short_name = "dmy"
        return OutputProduct(
            name=name,
            version=version,
            short_name=short_name,
            location=f"s3://dummy-bucket/{name}/{version}",
            properties={"odc:file_format": "GeoTIFF"},
            measurements=measurements,
        )


class WorkTokenInterface(ABC):
    @staticmethod
    def now():
        """
        Implementations should use this method internally
        """
        return datetime.utcnow()

    @property
    @abstractmethod
    def start_time(self) -> datetime:
        """
        Should return timestamp when task "started"
        """

    @property
    @abstractmethod
    def deadline(self) -> datetime:
        """
        Should return timestamp by which work is to be completed
        """

    @property
    @abstractmethod
    def done(self):
        """
        Called when work is completed successfully
        """

    @abstractmethod
    def cancel(self):
        """
        Called when work is terminated for whatever reason without successful result
        """
        pass

    @abstractmethod
    def extend(self, seconds: int) -> bool:
        """
        Called to extend work deadline
        """
        pass

    @property
    def active_seconds(self) -> float:
        """
        :returns: Number of seconds this Token has been active for
        """
        return (self.now() - self.start_time).total_seconds()

    def extend_if_needed(self, seconds, buffer_seconds: int = 30) -> bool:
        """
        Call ``.extend(seconds)`` only if deadline is within ``buffer_seconds`` from now

        :returns: True if deadline is still too far in the future
        :returns: Result of ``.extend(seconds)`` if deadline is close enough
        """
        t_now = self.now()
        if t_now + timedelta(seconds=buffer_seconds) > self.deadline:
            return self.extend(seconds)
        return True


@dataclass
class Task:
    product: OutputProduct
    tile_index: TileIdx_xy
    geobox: GeoBox
    time_range: DateTimeRange
    datasets: Tuple[Dataset, ...] = field(repr=False)
    uuid: UUID = UUID(int=0)
    short_time: str = field(init=False, repr=False)
    source: Optional[WorkTokenInterface] = field(init=True, repr=False, default=None)

    def __post_init__(self):
        self.short_time = self.time_range.short

        if self.uuid.int == 0:
            self.uuid = odc_uuid(
                self.product.name,
                self.product.version,
                sources=self._lineage(),
                time=self.short_time,
                tile=self.tile_index,
            )

    @property
    def location(self) -> str:
        """
        Product relative location for this task
        """
        rc = self.product.region_code(self.tile_index)
        mid = len(rc) // 2
        p1, p2 = rc[:mid], rc[mid:]
        return "/".join([p1, p2, self.short_time])

    def _lineage(self) -> Tuple[UUID, ...]:
        ds, *_ = self.datasets

        # TODO: replace this and test
        # if 'fused' in ds.metadata._doc['properties'].keys():
        if 'fused' in ds.type.name:
            lineage = tuple(set(
                x for ds in self.datasets for y in ds.metadata.sources.values() for x in y.values()
            ))
        else:
            lineage = tuple(ds.id for ds in self.datasets)

        return lineage

    def _prefix(self, relative_to: str = "dataset") -> str:
        product = self.product
        region_code = product.region_code(self.tile_index)
        file_prefix = f"{product.short_name}_{region_code}_{self.short_time}"
        parent_folder = f"{product.location}/{self.location}"

        # put maturity value: e.g. final as part of prefix
        maturity = self.product.maturity

        if relative_to == "dataset":
            return file_prefix if (maturity is None) \
                                    else file_prefix + "_" + maturity
        elif relative_to == "product":
            return self.location + "/" + file_prefix if (maturity is None) \
                                    else self.location + "/" + file_prefix + "_" + maturity
        else:
            return parent_folder + "/" + file_prefix if (maturity is None) \
                                     else parent_folder + "/" + file_prefix + "_" + maturity

    def paths(
        self, relative_to: str = "dataset", ext: str = EXT_TIFF
    ) -> Dict[str, str]:
        """
        Compute dictionary mapping band name to paths.

        :param relative_to: dataset|product|absolute
        """
        prefix = self._prefix(relative_to)
        return {band: f"{prefix}_{band}.{ext}" for band in self.product.measurements}

    def metadata_path(self, relative_to: str = "dataset", ext: str = "yaml") -> str:
        """
        Compute path for metadata file.

        :param relative_to: dataset|product|absolute
        """
        return self._prefix(relative_to) + "." + ext

    def aux_path(self, name: str, relative_to: str = "dataset", ext: str = EXT_TIFF):
        """
        Compute path for some auxilary file.

        :param relative_to: dataset|product|absolute
        :param name: "band"
        :param ext: File extension, defaults to tif
        """
        prefix = self._prefix(relative_to)
        return f"{prefix}_{name}.{ext}"

    def render_assembler_metadata(
        self, ext: str = EXT_TIFF, output_dataset: xr.Dataset = None, processing_dt: Optional[datetime] = None
    ) -> DatasetAssembler:
        """
        Put together metadata document for the output of this task. It needs the source_dataset to inherit
        several properties and lineages. It also needs the output_dataset to get the measurement information.
        """
        dataset_assembler = DatasetAssembler(naming_conventions=self.product.naming_conventions_values,
                                             dataset_location=Path(self.product.explorer_path),
                                             allow_absolute_paths=True)
        
        # ignore the tons of Inheritable property warnings
        warnings.simplefilter(action='ignore', category=UserWarning)

        platforms, instruments = ([], [])

        for dataset in self.datasets:
            if 'fused' in dataset.type.name:
                sources = [e['id'] for e in dataset.metadata.sources.values()]
                platforms.append(dataset.metadata_doc['properties']['eo:platform'])
                instruments.append(dataset.metadata_doc['properties']['eo:instrument'])
                dataset_assembler.note_source_datasets(self.product.classifier,
                                                       *sources)
            else:
                source_datasetdoc = serialise.from_doc(dataset.metadata_doc, skip_validation=True)
                dataset_assembler.add_source_dataset(source_datasetdoc,
                                                    classifier=self.product.classifier,
                                                    auto_inherit_properties=True, # it will grab all useful input dataset preperties
                                                    inherit_geometry=False,
                                                    inherit_skip_properties=self.product.inherit_skip_properties)

                if 'eo:platform' in source_datasetdoc.properties:
                    platforms.append(source_datasetdoc.properties['eo:platform'])
                if 'eo:instrument' in source_datasetdoc.properties:
                    instruments.append(source_datasetdoc.properties['eo:instrument'])

        dataset_assembler.platform = ','.join(sorted(set(platforms)))
        dataset_assembler.instrument = "_".join(sorted(set(instruments)))

        dataset_assembler.geometry = self.geobox.extent.geom

        dataset_assembler.datetime = format_datetime(self.time_range.start)
        dataset_assembler.properties["dtr:start_datetime"] = format_datetime(self.time_range.start)
        dataset_assembler.properties["dtr:end_datetime"] = format_datetime(self.time_range.end)

        # inherit properties from cfg
        for product_property_name, product_property_value in self.product.properties.items():
            dataset_assembler.properties[product_property_name] = product_property_value

        dataset_assembler.product_name = self.product.name
        dataset_assembler.dataset_version = self.product.version
        dataset_assembler.region_code = self.product.region_code(self.tile_index)

        # set the warning message back
        warnings.filterwarnings('default')

        if processing_dt is None:
            processing_dt = datetime.utcnow()
        dataset_assembler.processed = processing_dt

        dataset_assembler.maturity = self.product.maturity
        dataset_assembler.collection_number = self.product.collection_number

        if output_dataset is not None:
            for band, path in self.paths(ext=ext).items():
                # when we pass grid, the eodatasets will not load file from path
                dataset_assembler.note_measurement(band,
                                                    path,
                                                    expand_valid_data=False,
                                                    grid=GridSpec(shape=self.geobox.shape,
                                                                    transform=self.geobox.transform,
                                                                    crs=CRS.from_epsg(self.geobox.crs.to_epsg())),
                                                    nodata=output_dataset[band].nodata if 'nodata' in output_dataset[band].attrs else None)

        return dataset_assembler

    def render_metadata(
        self, ext: str = EXT_TIFF, processing_dt: Optional[datetime] = None
    ) -> Dict[str, Any]:
        """
        Put together STAC metadata document for the output of this task.
        """
        if processing_dt is None:
            processing_dt = datetime.utcnow()

        product = self.product
        geobox = self.geobox
        region_code = product.region_code(self.tile_index)
        inputs = list(map(str, self._lineage()))

        properties: Dict[str, Any] = deepcopy(product.properties)

        properties["dtr:start_datetime"] = format_datetime(self.time_range.start)
        properties["dtr:end_datetime"] = format_datetime(self.time_range.end)
        properties["odc:processing_datetime"] = format_datetime(
            processing_dt, timespec="seconds"
        )
        properties["odc:region_code"] = region_code
        properties["odc:product"] = product.name
        properties["odc:dataset_version"] = product.version

        geobox_wgs84 = geobox.extent.to_crs(
            "epsg:4326", resolution=math.inf, wrapdateline=True
        )
        bbox = geobox_wgs84.boundingbox

        item = pystac.Item(
            id=str(self.uuid),
            geometry=geobox_wgs84.json,
            bbox=[bbox.left, bbox.bottom, bbox.right, bbox.top],
            datetime=self.time_range.start.replace(tzinfo=timezone.utc),
            properties=properties
        )
        ProjectionExtension.add_to(item)
        proj_ext = ProjectionExtension.ext(item)
        proj_ext.apply(geobox.crs.epsg, transform=geobox.transform, shape=geobox.shape)

        # Lineage last
        item.properties["odc:lineage"] = dict(inputs=inputs)

        # Add all the assets
        for band, path in self.paths(ext=ext).items():
            asset = pystac.Asset(
                href=path,
                media_type="image/tiff; application=geotiff",
                roles=["data"],
                title=band,
            )
            item.add_asset(band, asset)

        # Add links
        item.links.append(
            pystac.Link(
                rel="product_overview",
                media_type="application/json",
                target=product.href,
            )
        )
        item.links.append(
            pystac.Link(
                rel="self",
                media_type="application/json",
                target=self.metadata_path("absolute", ext="json"),
            )
        )

        return item.to_dict()


def product_for_plugin(
    plugin: StatsPluginInterface,
    location: str,
    name: Optional[str] = None,
    short_name: Optional[str] = None,
    version: Optional[str] = None,
    product_family: Optional[str] = None,
    collections_site: str = "collections.dea.ga.gov.au",
    producer: str = "ga.gov.au",
    properties: Dict[str, Any] = dict(),
    region_code_format: str = "x{x:02d}y{y:02d}",
    naming_conventions_values: str = "dea_c3",
    explorer_path: str = "https://explorer.dea.ga.gov.au",
    inherit_skip_properties: Optional[List[str]] = None,
    preview_image_ows_style: Optional[Dict[str, Any]] = None,
    classifier: str = "level3",
    maturity: Optional[str] = None,
    collection_number: int = 3,
    nodata: Optional[Dict[str, int]] = None
) -> OutputProduct:
    """
    :param plugin: An instance of a subclass of StatsPluginInterface, used for name defaults.
    :param location: Output location string or template, example ``s3://bucket/{product}/{version}``
    :param name: Override for product name
    :param short_name: Override for product short_name
    :param version: Override for version
    :param product_family: Override for odc:product_family
    :param collections_site: href=f"https://{collections_site}/product/{name}"
    :param producer: Producer ``ga.gov.au``
    :param region_code_format: Change region code formatting, default ``"x{x:02d}y{y:02d}"``
    :param naming_conventions_values: default ``dea_c3``
    :param explorer_path: default ``https://explorer.dea.ga.gov.au``
    :param inherit_skip_properties: block properties from source datasets.
    :param preview_image_ows_style: define ows_styling_dict
    :param classifier: default ``level3``
    :param maturity: default ``None``
    :param collection_number: default ``3``
    :param nodata: band level nodata information. Pass it to eodatasets3 library only.
    """
    if name is None:
        name = plugin.NAME
    if short_name is None:
        short_name = plugin.SHORT_NAME
        if len(short_name) == 0:
            short_name = name
    if version is None:
        version = plugin.VERSION
    if product_family is None:
        product_family = plugin.PRODUCT_FAMILY

    if "{" in location and "}" in location:
        version_dashed = version.replace(".", "-")
        location = location.format(
            name=name,
            product=name,
            short_name=short_name,
            version=version_dashed,
            version_dashed=version_dashed,
            version_raw=version,
        )

    # remove trailing / if present
    location = location.rstrip("/")

    return OutputProduct(
        name=name,
        version=version,
        short_name=short_name,
        location=location,
        properties={
            "odc:file_format": "GeoTIFF",
            "odc:product_family": product_family,
            "odc:producer": producer,
            **properties,
        },
        measurements=plugin.measurements,
        href=f"https://{collections_site}/product/{name}",
        region_code_format=region_code_format,
        naming_conventions_values=naming_conventions_values,
        explorer_path=explorer_path,
        inherit_skip_properties=inherit_skip_properties,
        preview_image_ows_style=preview_image_ows_style,
        classifier=classifier,
        maturity=maturity,
        collection_number=collection_number,
        nodata=nodata,
    )


@dataclass
class TaskResult:
    task: Task
    result_location: str = ""
    skipped: bool = False
    error: Optional[str] = None
    meta: Any = field(init=True, repr=False, default=None)

    def __bool__(self):
        return self.error is None


@dataclass
class TaskRunnerConfig:
    @staticmethod
    def default_cog_settings():
        return dict(
            compress="deflate",
            zlevel=9,
            blocksize=800,
            ovr_blocksize=256,  # ovr_blocksize must be powers of 2 for some reason in GDAL
            overview_resampling="average",
        )

    # Input
    filedb: str = ""
    aws_unsigned: bool = True

    # Plugin
    plugin: str = ""
    plugin_config: Dict[str, Any] = field(init=True, repr=True, default_factory=dict)

    # Output Product
    #  .{name| short_name| version| product_family|
    #    collections_site| producer| properties: Dict[str, Any]}
    product: Dict[str, Any] = field(init=True, repr=True, default_factory=dict)

    # Dask config
    threads: int = -1
    memory_limit: str = ""

    # S3/Output config
    output_location: str = ""
    s3_acl: Optional[str] = None
    # s3_public is deprecated, use s3_acl="public-read" instead
    s3_public: bool = False
    cog_opts: Dict[str, Any] = field(init=True, repr=True, default_factory=dict)
    overwrite: bool = False

    # Heartbeat filepath
    heartbeat_filepath: Optional[str] = None

    # Terminate task if running longer than this amount (seconds)
    max_processing_time: int = 0

    # tuning/testing params
    #

    # SQS renew amount (seconds)
    job_queue_max_lease: int = 5 * 60

    # Renew work token when this close to deadline (seconds)
    renew_safety_margin: int = 30

    # How often future is checked for timeout/sqs renew
    future_poll_interval: float = 5

    def __post_init__(self):
        self.cog_opts = dicttoolz.merge(self.default_cog_settings(), self.cog_opts)


import logging
from typing import (
    Iterable,
    Iterator,
    Optional,
    List,
    Any,
    Tuple,
    Union,
)
from dask.distributed import Client
from datetime import datetime
import xarray as xr
import math
import psutil

from .model import Task, TaskResult, TaskRunnerConfig, product_for_plugin
from .io import S3COGSink
from ._text import read_int
from .tasks import TaskReader
from .plugins import resolve
from odc.algo import wait_for_future
from datacube.utils.dask import start_local_dask
from datacube.utils.rio import configure_s3_access


Future = Any


class TaskRunner:
    def __init__(
        self, cfg: TaskRunnerConfig, resolution: Optional[Tuple[float, float]] = None, from_sqs: Optional[str] = ""
    ):
        """

        """
        _log = logging.getLogger(__name__)
        self._cfg = cfg
        self._log = _log
        self.sink = S3COGSink(
            cog_opts=cfg.cog_opts, acl=cfg.s3_acl, public=cfg.s3_public
        )

        _log.info(f"Resolving plugin: {cfg.plugin}")
        mk_proc = resolve(cfg.plugin)
        self.proc = mk_proc(**cfg.plugin_config)
        self.product = product_for_plugin(self.proc, location=cfg.output_location, **cfg.product)
        _log.info(f"Output product: {self.product}")

        if not from_sqs:
            _log.info(f"Constructing task reader: {cfg.filedb}")
            self.rdr = TaskReader(cfg.filedb, self.product)
            _log.info(f"Will read from {self.rdr}")
            if resolution is not None:
                _log.info(f"Changing resolution to {resolution[0], resolution[1]}")
                if self.rdr.is_compatible_resolution(resolution):
                    self.rdr.change_resolution(resolution)
                else:
                    _log.error(
                        f"Requested resolution is not compatible with GridSpec in '{cfg.filedb}'"
                    )
                    raise ValueError(
                        f"Requested resolution is not compatible with GridSpec in '{cfg.filedb}'"
                    )
        else:  # skip rdr and resolution compatible init
            _log.info(f"Skip rdr init for run from sqs: {cfg.filedb}")
            self.rdr = TaskReader("", self.product, resolution)
            
        self._client = None

    def _init_dask(self) -> Client:
        cfg = self._cfg
        _log = self._log

        nthreads = cfg.threads
        if nthreads <= 0:
            nthreads = get_max_cpu()

        memory_limit: Union[str, int] = cfg.memory_limit
        if memory_limit == "":
            _1G = 1 << 30
            memory_limit = get_max_mem()
            if memory_limit > 2 * _1G:
                # leave at least a gig extra if total mem more than 2G
                memory_limit -= _1G

        client = start_local_dask(
            threads_per_worker=nthreads, processes=False, memory_limit=memory_limit
        )
        aws_unsigned = self._cfg.aws_unsigned
        for c in (None, client):
            configure_s3_access(
                aws_unsigned=aws_unsigned, cloud_defaults=True, client=c
            )
        _log.info(f"Started local Dask {client}")

        return client

    def client(self) -> Client:
        if self._client is None:
            self._client = self._init_dask()
        return self._client

    def verify_setup(self) -> bool:
        _log = self._log

        if self.product.location.startswith("s3://"):
            _log.info(f"Verifying credentials for output to {self.product.location}")

            test_uri = None  # TODO: for now just check credentials are present
            if not self.sink.verify_s3_credentials(test_uri):
                _log.error("Failed to obtain S3 credentials for writing")
                return False
        return True

    def tasks(self, tasks: List[str], ds_filters: Optional[str] = None,) -> Iterator[Task]:
        from ._cli_common import parse_all_tasks

        if len(tasks) == 0:
            tiles = self.rdr.all_tiles
        else:
            # this can throw ValueError
            tiles = parse_all_tasks(tasks, self.rdr.all_tiles)

        return self.rdr.stream(tiles, ds_filters=ds_filters)

    def dry_run(
        self, tasks: List[str], check_exists: bool = True, ds_filters: Optional[str] = None
    ) -> Iterator[TaskResult]:
        sink = self.sink
        overwrite = self._cfg.overwrite

        # exists (None|T|F) -> str
        flag_mapping = {
            None: "",
            False: " (new)",
            True: " (recompute)" if overwrite else " (skip)",
        }

        for task in self.tasks(tasks, ds_filters=ds_filters):
            uri = sink.uri(task)
            exists = None
            if check_exists:
                exists = sink.exists(task)

            skipped = (overwrite is False) and (exists is True)
            nds = len(task.datasets)
            # TODO: take care of utc offset for day boundaries when computing ndays
            ndays = len(set(ds.center_time.date() for ds in task.datasets))
            flag = flag_mapping.get(exists, "")
            msg = f"{task.location} days={ndays:03} ds={nds:04} {uri}{flag}"

            yield TaskResult(task, uri, skipped=skipped, meta=msg)

    def _safe_result(self, f: Future, task: Task) -> TaskResult:
        _log = self._log
        try:
            rr = f.result()
            if rr.error is None:
                return TaskResult(task, rr.path)
            else:
                error_msg = f"Failed to write: {rr.path}"
                _log.error(error_msg)
                return TaskResult(task, rr.path, error=error_msg)
        except Exception as e:
            _log.error(f"Error during processing of {task.location} {e}")
            return TaskResult(task, error=str(e))

    def _register_heartbeat(self, hearbeat_filepath: str):
        """
        Records the timestamp at which a hearbeat was detected

        """
        t_now = datetime.utcnow()
        with open(f"{hearbeat_filepath}", "w") as file_obj:
            file_obj.write(t_now.strftime("%Y-%m-%d %H:%M:%S"))

    def _run(self, tasks: Iterable[Task], apply_eodatasets3) -> Iterator[TaskResult]:
        cfg = self._cfg
        client = self.client()
        sink = self.sink
        proc = self.proc
        check_exists = cfg.overwrite is False
        _log = self._log

        for task in tasks:
            _log.info(f"Starting processing of {task.location}")
            tk = task.source
            if tk is not None:
                t0 = tk.start_time
            else:
                t0 = datetime.utcnow()
            if check_exists:
                path = sink.uri(task)
                _log.debug(f"Checking if can skip {path}")
                if sink.exists(task):
                    _log.info(f"Skipped task @ {path}")
                    if tk:
                        _log.info("Notifying completion via SQS")
                        tk.done()

                    yield TaskResult(task, path, skipped=True)
                    continue

            _log.debug("Building Dask Graph")
            ds = proc.reduce(proc.input_data(task.datasets, task.geobox))

            _log.debug(f"Submitting to Dask ({task.location})")
            ds = client.persist(ds, fifo_timeout="1ms")

            aux: Optional[xr.Dataset] = None

            # if no rgba setting in cog_ops:overrides, no rgba tif as ouput
            if 'overrides' in cfg.cog_opts and 'rgba' in cfg.cog_opts['overrides']: 
                rgba = proc.rgba(ds)
                if rgba is not None:
                    aux = xr.Dataset(dict(rgba=rgba))

            cog = sink.dump(task, ds, aux, proc, apply_eodatasets3)
            cog = client.compute(cog, fifo_timeout="1ms")

            _log.debug("Waiting for completion")
            cancelled = False

            for (dt, t_now) in wait_for_future(cog, cfg.future_poll_interval, t0=t0):
                if cfg.heartbeat_filepath is not None:
                    self._register_heartbeat(cfg.heartbeat_filepath)
                if tk:
                    tk.extend_if_needed(
                        cfg.job_queue_max_lease, cfg.renew_safety_margin
                    )
                if cfg.max_processing_time > 0 and dt > cfg.max_processing_time:
                    _log.error(
                        f"Task {task.location} failed to finish on time: {dt}>{cfg.max_processing_time}"
                    )
                    cancelled = True
                    cog.cancel()
                    break

            if cancelled:
                result = TaskResult(task, error="Cancelled due to timeout")
            else:
                result = self._safe_result(cog, task)

            if result:
                _log.info(f"Finished processing of {result.task.location}")
                if tk:
                    _log.info("Notifying completion via SQS")
                    tk.done()
            else:
                if tk:
                    tk.cancel()

            yield result

    def run(
        self, tasks: Optional[List[str]] = None, 
        sqs: Optional[str] = None, 
        ds_filters: Optional[str] = None,
        apply_eodatasets3: Optional[bool] = False
    ) -> Iterator[TaskResult]:
        cfg = self._cfg
        _log = self._log

        if tasks is not None:
            _log.info("Starting processing from task list")
            return self._run(self.tasks(tasks, ds_filters=ds_filters), apply_eodatasets3)
        if sqs is not None:
            _log.info(
                f"Processing from SQS: {sqs}, T:{cfg.job_queue_max_lease} M:{cfg.renew_safety_margin} seconds"
            )
            return self._run(
                self.rdr.stream_from_sqs(
                    sqs, visibility_timeout=cfg.job_queue_max_lease, ds_filters=ds_filters
                ), 
                apply_eodatasets3
            )
        raise ValueError("Must supply one of tasks= or sqs=")


def get_max_mem() -> int:
    """
    Max available memory, takes into account pod resource allocation
    """
    total = psutil.virtual_memory().total
    mem_quota = get_mem_quota()
    if mem_quota is None:
        return total
    return min(mem_quota, total)


def get_max_cpu() -> int:
    """
    Max available CPU (rounded up if fractional), takes into account pod
    resource allocation
    """
    ncpu = get_cpu_quota()
    if ncpu is not None:
        return int(math.ceil(ncpu))
    return psutil.cpu_count()


def get_cpu_quota() -> Optional[float]:
    """
    :returns: ``None`` if unconstrained or there is an error
    :returns: maximum amount of CPU this pod is allowed to use
    """
    quota = read_int("/sys/fs/cgroup/cpu/cpu.cfs_quota_us")
    if quota is None:
        return None
    period = read_int("/sys/fs/cgroup/cpu/cpu.cfs_period_us")
    if period is None:
        return None
    return quota / period


def get_mem_quota() -> Optional[int]:
    """
    :returns: ``None`` if there was some error
    :returns: maximum RAM, in bytes, this pod can use according to Linux cgroups

    Note that number returned can be larger than total available memory.
    """
    return read_int("/sys/fs/cgroup/memory/memory.limit_in_bytes")



"""
Various I/O adaptors
"""

from typing import Any, Dict, List, Optional, Union, cast
import json
from urllib.parse import urlparse
import logging
import dask
from dask.delayed import Delayed
from pathlib import Path
import xarray as xr
import io
from rasterio.crs import CRS

from datacube.utils.aws import get_creds_with_retry, mk_boto_session, s3_client
from odc.aws import s3_head_object  # TODO: move it to datacube
from datacube.utils.dask import save_blob_to_s3, save_blob_to_file
from datacube.utils.cog import to_cog
from datacube.model import Dataset
from botocore.credentials import ReadOnlyCredentials
from .model import Task, EXT_TIFF
from .plugins import StatsPluginInterface
from hashlib import sha1
from collections import namedtuple
from odc.stats import __version__

from eodatasets3.assemble import serialise
from eodatasets3.scripts.tostac import json_fallback
from eodatasets3.model import DatasetDoc
from eodatasets3.images import FileWrite, GridSpec
import eodatasets3.stac as eo3stac
import eodatasets3

WriteResult = namedtuple("WriteResult", ["path", "sha1", "error"])

_log = logging.getLogger(__name__)
DEFAULT_COG_OPTS = dict(compress="deflate", zlevel=6, blocksize=512,)


def load_creds(profile: Optional[str] = None) -> ReadOnlyCredentials:
    session = mk_boto_session(profile=profile)
    creds = get_creds_with_retry(session)
    if creds is None:
        raise ValueError("Failed to obtain credentials")

    return creds.get_frozen_credentials()


def dump_json(meta: Dict[str, Any]) -> str:
    return json.dumps(meta, separators=(",", ":"))


def mk_sha1(data):
    if isinstance(data, str):
        data = data.encode("utf8")
    return sha1(data).hexdigest()


_dask_sha1 = dask.delayed(mk_sha1, name="sha1")


@dask.delayed
def _pack_write_result(write_out, sha1):
    path, ok = write_out
    if ok:
        return WriteResult(path, sha1, None)
    else:
        return WriteResult(path, sha1, "Failed Write")


@dask.delayed(name="sha1-digest")
def _sha1_digest(*write_results):
    lines = []
    for wr in write_results:
        if wr.error is not None:
            raise IOError(f"Failed to write for: {wr.path}")
        file = wr.path.split("/")[-1]
        lines.append(f"{wr.sha1}\t{file}\n")
    return "".join(lines)


class S3COGSink:
    def __init__(
        self,
        creds: Union[ReadOnlyCredentials, str, None] = None,
        cog_opts: Optional[Dict[str, Any]] = None,
        acl: Optional[str] = None,
        public: bool = False,
    ):
        """
        :param creds: S3 write credentials
        :param cog_opts: Configure compression settings, globally and per-band
        :param acl: Canned ACL string:
                    private|public-read|public-read-write|authenticated-read|
                    aws-exec-read|bucket-owner-read|bucket-owner-full-control
        :param public: Mark objects as public access, same as `acl="public-read"`

        Example of COG config

        .. code-block:: python

           # - Lossless compression for all bands but rgba
           # - Webp compression for rgba band
           cfg = {
               "compression": "deflate",
               "zlevel": 9,
               "blocksize": 1024,
               "overrides": {"rgba": {"compression": "webp", "webp_level": 80}},
           }


        """

        if cog_opts is None:
            cog_opts = dict(**DEFAULT_COG_OPTS)
        else:
            tmp = dict(**DEFAULT_COG_OPTS)
            tmp.update(cog_opts)
            cog_opts = tmp

        cog_opts_per_band = cast(
            Dict[str, Dict[str, Any]], cog_opts.pop("overrides", {})
        )
        per_band_cfg = {k: v for k, v in cog_opts.items() if isinstance(v, dict)}
        if per_band_cfg:
            for k in per_band_cfg:
                cog_opts.pop(k)
            cog_opts_per_band.update(per_band_cfg)

        if acl is None and public:
            acl = "public-read"

        self._creds = creds
        self._cog_opts = cog_opts
        self._cog_opts_per_band = cog_opts_per_band
        self._stac_meta_ext = "stac-item.json"
        self._odc_meta_ext = "odc-metadata.yaml"
        self._proc_info_ext = "proc-info.yaml"
        self._stac_meta_contentype = "application/json"
        self._odc_meta_contentype = "text/yaml"
        self._prod_info_meta_contentype = "text/yaml"
        self._band_ext = EXT_TIFF
        self._acl = acl

    def uri(self, task: Task) -> str:
        return task.metadata_path("absolute", ext=self._stac_meta_ext)

    def _get_creds(self) -> ReadOnlyCredentials:
        if self._creds is None:
            self._creds = load_creds()
        if isinstance(self._creds, str):
            self._creds = load_creds(self._creds)
        return self._creds

    def verify_s3_credentials(self, test_uri: Optional[str] = None) -> bool:
        try:
            _ = self._get_creds()
        except ValueError:
            return False
        if test_uri is None:
            return True
        rr = self._write_blob(b"verifying S3 permissions", test_uri).compute()
        assert rr.path == test_uri
        return rr.error is None

    def _write_blob(
        self, data, url: str, ContentType: Optional[str] = None, with_deps=None
    ) -> Delayed:
        """
        Returns Delayed WriteResult[path, sha1, error=None]
        """
        _u = urlparse(url)
        sha1 = _dask_sha1(data)

        if _u.scheme == "s3":
            kw = dict(creds=self._get_creds())
            if ContentType is not None:
                kw["ContentType"] = ContentType
            if self._acl is not None:
                kw["ACL"] = self._acl

            return _pack_write_result(
                save_blob_to_s3(data, url, with_deps=with_deps, **kw), sha1
            )
        elif _u.scheme == "file":
            _dir = Path(_u.path).parent
            if not _dir.exists():
                _dir.mkdir(parents=True, exist_ok=True)
            return _pack_write_result(
                save_blob_to_file(data, _u.path, with_deps=with_deps), sha1
            )
        else:
            raise ValueError(f"Don't know how to save to '{url}'")

    def _ds_to_cog(self, ds: xr.Dataset, paths: Dict[str, str]) -> List[Delayed]:
        out = []
        for band, dv in ds.data_vars.items():
            band = str(band)
            url = paths.get(band, None)
            if url is None:
                raise ValueError(f"No path for band: '{band}'")
            cog_opts = self.cog_opts(band)
            cog_bytes = to_cog(dv, **cog_opts)
            out.append(self._write_blob(cog_bytes, url, ContentType="image/tiff"))
        return out

    def _apply_color_ramp(self, ds: xr.Dataset, ows_style_dict: dict, task: Task):
        from datacube_ows.styles.api import StandaloneStyle
        from datacube_ows.styles.api import apply_ows_style

        ows_style = StandaloneStyle(ows_style_dict)
        # assign the time to xr.Dataset cause ows needs it
        time_da = xr.Dataset({"time": task.time_range.start})
        dst = ds.expand_dims(time=time_da.to_array())
        return apply_ows_style(ows_style, dst)


    def _get_thumbnail(self, ds: xr.Dataset, task: Task, ows_style_dict: dict, input_geobox: GridSpec, odc_file_path: str) -> Delayed:
        display_pixels = []

        image = self._apply_color_ramp(ds, ows_style_dict, task)
        # apply the OWS styling, the return image must have red, green, blue and alpha.
        for display_band in ['red', 'green', 'blue']:
            display_pixels.append(image[display_band].values.reshape([task.geobox.shape[0], task.geobox.shape[1]]))

        thumbnail_bytes = FileWrite().create_thumbnail_from_numpy(rgb=display_pixels,
                                                                  static_stretch=(0, 255),
                                                                  out_scale=10,
                                                                  input_geobox=input_geobox,
                                                                  nodata=None)

        return self._write_blob(thumbnail_bytes, odc_file_path.split('.')[0] + f"_thumbnail.jpg", ContentType="image/jpeg")

    def _ds_to_thumbnail_cog(self, ds: xr.Dataset, task: Task) -> List[Delayed]:
        odc_file_path = task.metadata_path("absolute", ext=self._odc_meta_ext)

        thumbnail_cogs = []

        input_geobox = GridSpec(shape=task.geobox.shape, 
                                transform=task.geobox.transform, 
                                crs=CRS.from_epsg(task.geobox.crs.to_epsg()))

        if task.product.preview_image_ows_style:
            try:
                thumbnail_cog = self._get_thumbnail(ds, task, task.product.preview_image_ows_style, input_geobox, odc_file_path)
                thumbnail_cogs.append(thumbnail_cog)
            except AttributeError:
                _log.error(f"Cannot parse OWS styling: {task.product.preview_image_ows_style}.")
            except ImportError as e:
                raise type(e)(str(e) + '. Please run python -m pip install "odc-stats[ows]" to setup environment to generate thumbnail.')

        return thumbnail_cogs

    def cog_opts(self, band_name: str = "") -> Dict[str, Any]:
        opts = dict(self._cog_opts)
        opts.update(self._cog_opts_per_band.get(band_name, {}))
        return opts

    def write_cog(self, da: xr.DataArray, url: str) -> Delayed:
        cog_bytes = to_cog(da, **self.cog_opts(str(da.name)))
        return self._write_blob(cog_bytes, url, ContentType="image/tiff")

    def exists(self, task: Union[Task, str]) -> bool:
        if isinstance(task, str):
            uri = task
        else:
            uri = self.uri(task)
        _u = urlparse(uri)
        if _u.scheme == "s3":
            s3 = s3_client(creds=self._get_creds(), cache=True)
            meta = s3_head_object(uri, s3=s3)
            return meta is not None
        elif _u.scheme == "file":
            return Path(_u.path).exists()
        else:
            raise ValueError(f"Can't handle url: {uri}")

    def get_eo3_stac_meta(self, task: Task, meta: DatasetDoc, stac_file_path: str, odc_file_path: str) -> str:
        """
        Convert the eodatasets3 DatasetDoc to stac meta format string.
        The stac_meta is Python dict, please use json_fallback() to format it. Also pass dataset_location
        to convert all accessories to full url. The S3 and local dir will use different ways to extract.
        """
        _u = urlparse(stac_file_path)

        if _u.scheme == "s3":
            dataset_location = f"{_u.scheme}://{_u.netloc}/{'/'.join(_u.path.split('/')[:-1])}"
        else:
            dataset_location = str(Path(_u.path).parent)

        stac_meta = eo3stac.to_stac_item(dataset=meta,
                                        stac_item_destination_url=stac_file_path,
                                        dataset_location=dataset_location,
                                        odc_dataset_metadata_url =odc_file_path,
                                        explorer_base_url = task.product.explorer_path
                                        )
        return json.dumps(stac_meta, default=json_fallback) # stac_meta is Python str, but content is 'Dict format'

    def dump_with_pystac(self, task: Task, ds: Dataset, aux: Optional[Dataset] = None) -> Delayed:
        """
        Dump files with STAC metadata file, which generated from PySTAC
        """
        json_url = task.metadata_path("absolute", ext=self._stac_meta_ext)
        meta = task.render_metadata(ext=self._band_ext)
        json_data = dump_json(meta).encode("utf8")

        # fake write result for metadata output, we want metadata file to be
        # the last file written, so need to delay it until after sha1 is
        # written.
        meta_sha1 = dask.delayed(WriteResult(json_url, mk_sha1(json_data), None))

        paths = task.paths("absolute", ext=self._band_ext)
        cogs = self._ds_to_cog(ds, paths)

        if aux is not None:
            aux_paths = {
                k: task.aux_path(k, relative_to="absolute", ext=self._band_ext)
                for k in aux.data_vars
            }
            cogs.extend(self._ds_to_cog(aux, aux_paths))

        # this will raise IOError if any write failed, hence preventing json
        # from being written
        sha1_digest = _sha1_digest(meta_sha1, *cogs)
        sha1_url = task.metadata_path("absolute", ext="sha1")
        sha1_done = self._write_blob(sha1_digest, sha1_url, ContentType="text/plain")

        return self._write_blob(
            json_data, json_url, ContentType=self._stac_meta_contentype, with_deps=sha1_done,
        )

    def dump_with_eodatasets3(self, task: Task, ds: Dataset, aux: Optional[Dataset] = None, proc: StatsPluginInterface = None) -> Delayed:
        """
        Dump files with metadata files, which generated from eodatasets3
        """
        stac_file_path = task.metadata_path("absolute", ext=self._stac_meta_ext)
        odc_file_path = task.metadata_path("absolute", ext=self._odc_meta_ext)
        sha1_url = task.metadata_path("absolute", ext="sha1")
        proc_info_url = task.metadata_path("absolute", ext=self._proc_info_ext)
        dataset_assembler = task.render_assembler_metadata(ext=self._band_ext, output_dataset=ds)

        dataset_assembler.extend_user_metadata("input-products", sorted(set([e.type.name for e in task.datasets])))

        dataset_assembler.extend_user_metadata("odc-stats-config", vars(task.product))

        dataset_assembler.note_software_version("eodatasets3",
                                                "https://github.com/GeoscienceAustralia/eo-datasets",
                                                eodatasets3.__version__,)

        dataset_assembler.note_software_version('odc-stats',
                                                "https://github.com/opendatacube/odc-tools",
                                                __version__)

        dataset_assembler.note_software_version(proc.NAME,
                                                "https://github.com/opendatacube/odc-tools",
                                                proc.VERSION)

        if task.product.preview_image_ows_style:
            try:
                import datacube_ows
                thumbnail_path = odc_file_path.split('.')[0] + f"_thumbnail.jpg"
                dataset_assembler._accessories["thumbnail"] = Path(urlparse(thumbnail_path).path).name

                dataset_assembler.note_software_version("datacube-ows",
                                                        "https://github.com/opendatacube/datacube-ows",
                                                        # Just realized the odc-stats does not have version.
                                                        datacube_ows.__version__)
            except ImportError as e:
                raise type(e)(str(e) + '. Please run python -m pip install "odc-stats[ows]" to setup environment to generate thumbnail.')

        dataset_assembler._accessories["checksum:sha1"] = Path(urlparse(sha1_url).path).name
        dataset_assembler._accessories["metadata:processor"] = Path(urlparse(proc_info_url).path).name

        meta = dataset_assembler.to_dataset_doc()
        # already add all information to dataset_assembler, now convert to odc and stac metadata format

        stac_meta = self.get_eo3_stac_meta(task, meta, stac_file_path, odc_file_path)

        odc_meta_stream = io.StringIO("") # too short, not worth to move to another method.
        serialise.to_stream(odc_meta_stream, meta)
        odc_meta = odc_meta_stream.getvalue() # odc_meta is Python str

        proc_info_meta_stream = io.StringIO("")
        serialise._init_yaml().dump({**dataset_assembler._user_metadata, "software_versions": dataset_assembler._software_versions}, proc_info_meta_stream)
        proc_info_meta = proc_info_meta_stream.getvalue()

        # fake write result for metadata output, we want metadata file to be
        # the last file written, so need to delay it until after sha1 files is
        # written.
        stac_meta_sha1 = dask.delayed(WriteResult(stac_file_path, mk_sha1(stac_meta), None))
        odc_meta_sha1 = dask.delayed(WriteResult(odc_file_path, mk_sha1(odc_meta), None))
        proc_info_sha1 = dask.delayed(WriteResult(proc_info_url, mk_sha1(proc_info_meta), None))

        paths = task.paths("absolute", ext=self._band_ext)
        cogs = self._ds_to_cog(ds, paths)

        if aux is not None:
            aux_paths = {
                k: task.aux_path(k, relative_to="absolute", ext=self._band_ext)
                for k in aux.data_vars
            }
            cogs.extend(self._ds_to_cog(aux, aux_paths))

        thumbnail_cogs = self._ds_to_thumbnail_cog(ds, task)

        # this will raise IOError if any write failed, hence preventing json
        # from being written
        sha1_digest = _sha1_digest(stac_meta_sha1, odc_meta_sha1, proc_info_sha1, *cogs, *thumbnail_cogs)
        sha1_done = self._write_blob(sha1_digest, sha1_url, ContentType="text/plain")

        proc_info_done = self._write_blob(proc_info_meta, proc_info_url, ContentType=self._prod_info_meta_contentype, with_deps=sha1_done)
        odc_meta_done = self._write_blob(odc_meta, odc_file_path, ContentType=self._odc_meta_contentype, with_deps=proc_info_done)
        cog_done = self._write_blob(stac_meta, stac_file_path, ContentType=self._stac_meta_contentype, with_deps=odc_meta_done)

        # The uploading DAG is:
        # sha1_done -> proc_info_done -> odc_meta_done -> stac_meta_done
        return cog_done

    def dump(self, task: Task, ds: Dataset, aux: Optional[Dataset] = None, proc: StatsPluginInterface = None, apply_eodatasets3: Optional[bool] = False) -> Delayed:
        
        if apply_eodatasets3:
            return self.dump_with_eodatasets3(task, ds, aux, proc)
        else:
            return self.dump_with_pystac(task, ds, aux)


import click
import itertools
from ._cli_common import main

from odc.dscache import create_cache
from ._stac_fetch import s3_fetch_dss


@main.command("generate-cache")
@click.argument("input_glob", type=str)
@click.argument("location", type=str)
@click.option("--verbose", "-v", is_flag=True, help="Be verbose")
def cli(input_glob, location, verbose):
    """
    Dump stats data into a cache file.

    Note: The input bucket must be public otherwise the data can not be listed.
    """

    # Look ahead to get product
    dss = s3_fetch_dss(input_glob)
    ds0 = next(dss)
    product = ds0.type
    dss = itertools.chain(iter([ds0]), dss)

    cache = create_cache(f"{product.name}.db")

    if verbose:
        print(f"Writing {location}/{product.name}.db")

    cache = create_cache(f"{location}/{product.name}.db")
    cache.bulk_save(dss)
    if verbose:
        print(f"Found {cache.count:,d} datasets")


if __name__ == "__main__":
    cli()



"""
Sentinel 2 pixel quality stats
"""
from typing import Dict, Iterable, Optional, Tuple, cast

import xarray as xr
from odc.algo import enum_to_bool, mask_cleanup
from odc.algo._masking import _or_fuser

from ._registry import StatsPluginInterface, register

cloud_classes = (
    "cloud shadows",
    "cloud medium probability",
    "cloud high probability",
    "thin cirrus",
)

# filters - dict of band name and list of iterable tuples of morphological operations
#           in the order you want them to perform
default_filters = {
  "clear_2_5": [("opening", 2), ("dilation", 5)],
  "clear_0_5": [("opening", 0), ("dilation", 5)]
}


class StatsPQ(StatsPluginInterface):
    NAME = "pc_s2_annual"
    SHORT_NAME = NAME
    VERSION = '0.0.1'
    PRODUCT_FAMILY = "pixel_quality_statistics"

    def __init__(
        self,
        filters: Optional[Dict[str, Iterable[Tuple[str, int]]]] = None,
        resampling: str = "nearest",
        **kwargs
    ):
        super().__init__(input_bands=["SCL"], resampling=resampling, **kwargs)
        if filters is None:
            filters = default_filters
        self.filters = filters

    @property
    def measurements(self) -> Tuple[str, ...]:
        measurements = [
            "total",
            "clear",
            *list(self.filters)
        ]

        return tuple(measurements)

    def fuser(self, xx: xr.Dataset) -> xr.Dataset:
        """
        Native:
          .valid    -> .valid  (fuse with OR)
          .erased   -> .erased (fuse with OR)
                    -> .erased{postfix} for All Filters (mask_cleanup applied post-fusing)
        Projected:
         fuse all bands with OR

        """
        is_native = xx.attrs.get("native", False)
        xx = xx.map(_or_fuser)
        xx.attrs.pop("native", None)

        if is_native:
            for band, mask_filters in self.filters.items():
                erased_filter_band_name = band.replace("clear", "erased")
                xx[erased_filter_band_name] = mask_cleanup(xx["erased"], mask_filters=mask_filters)

        return xx

    def native_transform(self, xx: xr.Dataset) -> xr.Dataset:
        """
        config:
        cloud_classes

        .SCL -> .valid   (anything but nodata)
                .erased  (cloud pixels only)
        """
        scl = xx.SCL
        valid = scl != scl.nodata
        erased = enum_to_bool(scl, cloud_classes)
        return xr.Dataset(
            dict(valid=valid, erased=erased),
            attrs={"native": True},  # <- native flag needed for fuser
        )

    def reduce(self, xx: xr.Dataset) -> xr.Dataset:
        """
        .valid            -> .total
        .erased           -> .clear
        .erased{postfix}  -> .clear{postfix} For All {postfix}
        """
        valid = xx.valid
        erased_bands = [str(n) for n in xx.data_vars if str(n).startswith("erased")]
        total = valid.sum(axis=0, dtype="uint16")
        pq = xr.Dataset(dict(total=total))

        for band in erased_bands:
            erased: xr.DataArray = cast(xr.DataArray, xx[band])
            clear_name = band.replace("erased", "clear")
            clear = valid * (~erased)  # type: ignore
            pq[clear_name] = clear.sum(axis=0, dtype="uint16")

        return pq


register("pq", StatsPQ)



import sys
import click
import json

from ._cli_common import main, parse_all_tasks


def do_dry_run(tasks):
    for period, ix, iy in tasks:
        print(f"{period}/{ix:+04d}/{iy:+04d}")


@main.command("publish-tasks")
@click.argument("db", type=str)
@click.argument("queue", type=str)
@click.option("--verbose", "-v", is_flag=True, help="Be verbose")
@click.option(
    "--dryrun", is_flag=True, help="Do not publish just print what would be submitted"
)
@click.option(
    "--bunch-size", type=int, default=10, help="Number of messages to submit in one go"
)
@click.argument("tasks", type=str, nargs=-1)
def publish_to_queue(db, queue, verbose, dryrun, bunch_size, tasks):
    """
    Publish tasks to SQS.

    Task could be one of the 3 things

    \b
    1. Comma-separated triplet: period,x,y or 'x[+-]<int>/y[+-]<int>/period
       2019--P1Y,+003,-004
       2019--P1Y/3/-4          `/` is also accepted
       x+003/y-004/2019--P1Y   is accepted as well
    2. A zero based index
    3. A slice following python convention <start>:<stop>[:<step]
        ::10 -- every tenth task: 0,10,20,..
       1::10 -- every tenth but skip first one 1, 11, 21 ..
        :100 -- first 100 tasks

    If no tasks are supplied all tasks will be published the queue.
    """
    from odc.aws.queue import get_queue, publish_messages
    from .tasks import TaskReader, render_sqs
    import toolz

    rdr = TaskReader(db)
    if len(tasks) == 0:
        tasks = rdr.all_tiles
        if verbose:
            print(f"Found {len(tasks):,d} tasks in the file")
    else:
        try:
            tasks = parse_all_tasks(tasks, rdr.all_tiles)
        except ValueError as e:
            print(str(e), file=sys.stderr)
            sys.exit(1)

    if dryrun:
        do_dry_run(tasks)
        sys.exit(0)

    queue = get_queue(queue)
    
    # We assume the db files are always be the S3 uri. If they are not, there is no need to use SQS queue to process.
    messages = (
        dict(Id=str(idx), MessageBody=json.dumps(render_sqs(tidx, db)))
        for idx, tidx in enumerate(tasks)
    )

    for bunch in toolz.partition_all(bunch_size, messages):
        publish_messages(queue, bunch)



"""
Work token for SQS based job control
"""
from typing import Optional
from datetime import timedelta, datetime
import toolz
from .model import WorkTokenInterface


class SQSWorkToken(WorkTokenInterface):
    def __init__(self, msg, timeout: int, t0: Optional[datetime] = None):
        if t0 is None:
            t0 = self.now()
        self._msg = msg
        self._t0 = t0
        self._deadline = t0 + timedelta(seconds=timeout)

    @property
    def start_time(self) -> datetime:
        """
        Timestamp when SQS message was received
        """
        return self._t0

    @property
    def deadline(self) -> datetime:
        """
        Should return timestamp by which work is to be completed
        """
        return self._deadline

    def done(self):
        """
        Called when work is completed successfully
        """
        if self._msg is not None:
            self._msg.delete()
            self._msg = None

    def cancel(self):
        """
        Called when work is terminated for whatever reason without successful result
        """
        if self._msg is None:
            return

        self._msg.change_visibility(VisibilityTimeout=0)
        self._msg = None

    def extend(self, seconds: int) -> bool:
        """
        Called to extend work deadline
        """
        if self._msg is None:
            return False

        new_deadline = self.now() + timedelta(seconds=seconds)

        rr = self._msg.change_visibility(VisibilityTimeout=seconds)
        ok = toolz.get_in(["ResponseMetadata", "HTTPStatusCode"], rr, default=-1) == 200

        if ok:
            self._deadline = new_deadline

        return ok



from typing import List, Tuple
import click

from ._text import parse_yaml_file_or_inline, parse_range2d_int

TileIdx_txy = Tuple[str, int, int]


def parse_task(s: str) -> TileIdx_txy:
    """
    Intentional copy of tasks.parse_task only for CLI parsing
    """
    from ._text import split_and_check

    sep = "/" if "/" in s else ","
    t, x, y = split_and_check(s, sep, 3)
    if t.startswith("x"):
        t, x, y = y, t, x
    return (t, int(x.lstrip("x")), int(y.lstrip("y")))


def parse_all_tasks(
    inputs: List[str], all_possible_tasks: List[TileIdx_txy]
) -> List[TileIdx_txy]:
    """
    Select a subset of all possible tasks given user input on cli.

    Every input task can be one of:
     <int>                   -- 0 based index into all_possible_tasks list

     <start>:<stop>[:<step>] -- slice of all_possible_tasks, 1:100, ::10, 1::100, -10:

     t,x,y or x,y,t triplet  -- Task index as a triplet of period,x,y
       2019--P1Y/10/-3
       2019--P1Y,10,-3
       x+10/y-3/2019--P1Y
    """
    from ._text import parse_slice

    out: List[TileIdx_txy] = []
    full_set = set(all_possible_tasks)

    for s in inputs:
        if "," in s or "/" in s:
            task = parse_task(s)
            if task not in full_set:
                raise ValueError(f"No task matches '{s}'")
            out.append(task)
        elif ":" in s:
            ii = parse_slice(s)
            out.extend(all_possible_tasks[ii])
        else:
            try:
                idx = int(s)
            except ValueError:
                raise ValueError(f"Failed to parse '{s}'") from None

            if idx < 0 or idx >= len(all_possible_tasks):
                raise ValueError(f"Task index is out of range: {idx}")
            out.append(all_possible_tasks[idx])

    return out


def parse_resolution(s: str, separator: str = ",") -> Tuple[float, float]:
    from ._text import split_and_check

    parts = [float(v) for v in split_and_check(s, separator, (1, 2))]

    if len(parts) == 1:
        return (-parts[0], parts[0])

    return (parts[0], parts[1])


def click_resolution(*args, **kw):
    """
    @click_resolution("--custom-flag-for-resolution", help="Whatever help")
    """
    def _parse(ctx, param, value):
        if value is not None:
            try:
                return parse_resolution(value)
            except ValueError as e:
                raise click.ClickException(str(e)) from None
    if len(args) == 0:
        args = ["--resolution"]
    return click.option(*args, callback=_parse, **kw)


def click_yaml_cfg(*args, **kw):
    """
    @click_yaml_cfg("--custom-flag", help="Whatever help")
    """
    def _parse(ctx, param, value):
        if value is not None:
            try:
                return parse_yaml_file_or_inline(value)
            except Exception as e:
                raise click.ClickException(str(e)) from None
    return click.option(*args, callback=_parse, **kw)


def setup_logging(level: int = -1):
    """
    Setup logging to print to stdout with default logging level being INFO.
    """
    import logging
    import sys

    if level < 0:
        level = logging.INFO

    logging.basicConfig(
        level=level,
        format="[%(asctime)s] {%(filename)s:%(lineno)d} %(levelname)s - %(message)s",
        stream=sys.stdout,
    )


def click_range2d(ctx, param, value):
    """
    @click.option('--range', callback=click_range2d)
    """
    if value is not None:
        try:
            return parse_range2d_int(value)
        except ValueError as e:
            raise click.ClickException(str(e)) from None


# pylint: disable=import-outside-toplevel,inconsistent-return-statements


@click.group(help="Stats command line interface")
def main():
    pass



from pathlib import Path
from typing import Union, Optional, Tuple, Dict, Any

PathLike = Union[str, Path]


# Copied from odc.io.text

def read_int(path: PathLike, default=None, base=10) -> Optional[int]:
    """
    Read single integer from a text file.

    Useful for things like parsing content of /sys/ or /proc.
    """
    try:
        with open(path, "rt") as f:
            return int(f.read(), base)
    except (FileNotFoundError, ValueError):
        return default


def split_and_check(
        s: str, separator: str, n: Union[int, Tuple[int, ...]]
) -> Tuple[str, ...]:
    """Turn string into tuple, checking that there are exactly as many parts as expected.
    :param s: String to parse
    :param separator: Separator character
    :param n: Expected number of parts, can be a single integer value or several,
              example `(2, 3)` accepts 2 or 3 parts.
    """
    if isinstance(n, int):
        n = (n,)

    parts = s.split(separator)
    if len(parts) not in n:
        raise ValueError('Failed to parse "{}"'.format(s))
    return tuple(parts)


def parse_slice(s: str) -> slice:
    """
    Parse slice syntax in the form start:stop[:step]
    Examples "::4", "2:5", "2::10", "3:100:5"
    """

    def parse(part: str) -> Optional[int]:
        if part == "":
            return None
        return int(part)

    try:
        parts = [parse(p) for p in split_and_check(s, ":", (2, 3))]
    except ValueError:
        raise ValueError(f'Expect <start>:<stop>[:<step>] syntax, got "{s}"') from None

    return slice(*parts)


def parse_yaml(s: str) -> Dict[str, Any]:
    # pylint: disable=import-outside-toplevel
    import yaml

    return yaml.load(s, Loader=getattr(yaml, "CSafeLoader", yaml.SafeLoader))


def parse_yaml_file_or_inline(s: str) -> Dict[str, Any]:
    """
    Accept on input either a path to yaml file or yaml text, return parsed yaml document.
    """
    try:
        # if file
        path = Path(s)
        with open(path, "rt") as f:
            txt = f.read()
            assert isinstance(txt, str)
    except (FileNotFoundError, IOError, ValueError):
        txt = s
    result = parse_yaml(txt)
    if isinstance(result, str):
        raise IOError(f"No such file: {s}")
    return result


def parse_range2d_int(s: str) -> Tuple[Tuple[int, int], Tuple[int, int]]:
    """Parse string like "0:3,4:5" -> ((0,3), (4,5))"""
    from ._text import split_and_check
    try:
        return tuple(tuple(int(x) for x in split_and_check(p, ":", 2)) for p in split_and_check(s, ",", 2))
    except ValueError:
        raise ValueError(
            'Expect <int>:<int>,<int>:<int> syntax, got "{}"'.format(s)
        ) from None


"""
Landsat QA Pixel Geomedian
"""
from functools import partial
from typing import Any, Dict, Iterable, Optional, Sequence, Tuple

import xarray as xr
import numpy as np
from datacube.utils import masking
from odc.algo import geomedian_with_mads, keep_good_only, erase_bad
from odc.algo._masking import _xr_fuse, _first_valid_np, mask_cleanup, _fuse_or_np
from ._registry import StatsPluginInterface, register

class StatsGMLSBitmask(StatsPluginInterface):
    NAME = "gm_ls_bitmask"
    SHORT_NAME = NAME
    VERSION = "0.0.1"

    def __init__(
            self,
            bands: Optional[Sequence[str]] = None,
            mask_band: str = "QA_PIXEL",
            # provide flags with high cloud bits definition
            flags: Dict[str, Optional[Any]] = dict(
                cloud="high_confidence",
                cirrus="high_confidence",
            ),
            nodata_flags: Dict[str, Optional[Any]] = dict(nodata=False),
            filters: Optional[Iterable[Tuple[str, int]]] = None, # e.g. [("closing", 10),("opening", 2),("dilation", 2)]
            aux_names=dict(smad="smad", emad="emad", bcmad="bcmad", count="count"),
            work_chunks: Tuple[int, int] = (400, 400),
            scale: float = 0.0000275,
            offset: float = -0.2,
            output_scale: int = 10000, # gm rescaling - making SR range match sentinel-2 gm
            output_dtype: str = "uint16", # dtype of gm rescaling
            **kwargs,
    ):
        if bands is None:
            self.bands = (
                "red",
                "green",
                "blue",
                "nir",
                "swir_1",
                "swir_2",
            )
        else:
            self.bands = bands
        self.mask_band = mask_band
        super().__init__(input_bands=tuple(bands) + (mask_band,), **kwargs)
        self.flags = flags
        self.nodata_flags = nodata_flags
        self.filters = filters
        self.work_chunks = work_chunks
        self.renames = aux_names
        self.aux_bands = tuple(aux_names.values())
        self.scale = scale
        self.offset = offset
        self.output_scale = output_scale
        self.output_dtype = np.dtype(output_dtype)
        self.output_nodata = 0

    @property
    def measurements(self) -> Tuple[str, ...]:
        return self.bands + self.aux_bands

    def native_transform(self, xx):
        """
        Loads in the data in the native projection. It performs the following:

        1. Loads pq bands
        2. Extract valid - by removing negative pixel using masking_scale from input bands
        3. Extract cloud_mask flags from bands
        4. Drop nodata and negative pixels
        5. Add cloud_mask band to xx for fuser and reduce

        .. bitmask::
            15  14  13  12  11  10  9   8   7   6   5   4   3   2   1   0
            |   |   |   |   |   |   |   |   |   |   |   |   |   |   |   |
            |   |   |   |   |   |   |   |   |   |   |   |   |   |   |   x-----> nodata
            |   |   |   |   |   |   |   |   |   |   |   |   |   |   o---------> dilated_cloud
            |   |   |   |   |   |   |   |   |   |   |   |   |   x-------------> cirrus
            |   |   |   |   |   |   |   |   |   |   |   |   o-----------------> cloud
            |   |   |   |   |   |   |   |   |   |   |   x---------------------> cloud_shadow
            |   |   |   |   |   |   |   |   |   |   o-------------------------> snow
            |   |   |   |   |   |   |   |   |   x-----------------------------> clear
            |   |   |   |   |   |   |   |   o---------------------------------> water
            |   |   |   |   |   |   |   x-------------------------------------> cloud_confidence
            |   |   |   |   |   |   o-----------------------------------------> cloud_confidence
            |   |   |   |   |   x---------------------------------------------> cloud_shadow_confidence
            |   |   |   |   o-------------------------------------------------> cloud_shadow_confidence
            |   |   |   x-----------------------------------------------------> snow_ice_confidence
            |   |   o---------------------------------------------------------> snow_ice_confidence
            |   x-------------------------------------------------------------> cirrus_confidence
            0-----------------------------------------------------------------> cirrus_confidence
        """

        # remove negative pixels - a pixel is invalid if any of the band is smaller than masking_scale
        valid = (xx[self.bands] > (-1.0 * self.offset/self.scale)).to_array(dim='band').all(dim='band')

        mask_band = xx[self.mask_band]
        xx = xx.drop_vars([self.mask_band])

        flags_def = masking.get_flags_def(mask_band)

        # set cloud_mask - True=cloud, False=non-cloud
        mask, _ = masking.create_mask_value(flags_def, **self.flags)
        cloud_mask = (mask_band & mask) != 0

        # set no_data bitmask - True=data, False=no-data
        nodata_mask, _ = masking.create_mask_value(flags_def, **self.nodata_flags)
        keeps = (mask_band & nodata_mask) == 0

        xx = keep_good_only(xx, valid)   # remove negative pixels
        xx = keep_good_only(xx, keeps)   # remove nodata pixels
        xx["cloud_mask"] = cloud_mask

        return xx

    def reduce(self, xx: xr.Dataset) -> xr.Dataset:
        cloud_mask = xx["cloud_mask"]
        cfg = dict(
            maxiters=1000,
            num_threads=1,
            scale=self.scale,
            offset=self.offset,
            reshape_strategy="mem",
            out_chunks=(-1, -1, -1),
            work_chunks=self.work_chunks,
            compute_count=True,
            compute_mads=True,
        )

        if self.filters is not None:
            cloud_mask = mask_cleanup(xx["cloud_mask"], mask_filters=self.filters)

        # erase pixels with cloud
        xx = xx.drop_vars(["cloud_mask"])
        xx = erase_bad(xx, cloud_mask)

        gm = geomedian_with_mads(xx, **cfg)
        gm = gm.rename(self.renames)

        # rescale gm bands into surface reflectance scale
        for band in gm.data_vars.keys():
            if band in self.bands or band == self.renames['emad']:
                # set nodata_mask - use for resetting nodata pixel after rescale
                nodata_mask = gm[band] == gm[band].attrs.get('nodata')
                # rescale
                if band == self.renames['emad']:
                    gm[band] = self.scale * self.output_scale * gm[band]
                else:
                    gm[band] = self.scale * self.output_scale * gm[band] + self.offset * self.output_scale
                #  apply nodata_mask - reset nodata pixels to output-nodata
                gm[band] = gm[band].where(~nodata_mask, self.output_nodata)
                # set data-type and nodata attrs
                gm[band] = gm[band].round().astype(self.output_dtype)
                gm[band].attrs['nodata'] = self.output_nodata

        return gm

    def fuser(self, xx):
        """
        Fuse cloud_mask with OR
        """
        cloud_mask = xx["cloud_mask"]
        xx = _xr_fuse(xx.drop_vars(["cloud_mask"]), partial(_first_valid_np, nodata=0), '')
        xx["cloud_mask"] = _xr_fuse(cloud_mask, _fuse_or_np, cloud_mask.name)

        return xx


register("gm-ls-bitmask", StatsGMLSBitmask)



import json

import click
import sys
from ._cli_common import main, click_range2d
from .utils import fuse_products, fuse_ds
from odc.index import ordered_dss, dataset_count
from itertools import groupby


@main.command("save-tasks")
@click.option(
    "--grid",
    type=str,
    help=(
        "Grid name or spec: au-{10|20|30|60},africa-{10|20|30|60}, albers-au-25 (legacy one)"
        "'crs;pixel_resolution;shape_in_pixels'"
    ),
    prompt="""Enter GridSpec
 one of au-{10|20|30|60}, africa-{10|20|30|60}, albers_au_25 (legacy one)
 or custom like 'epsg:3857;30;5000' (30m pixels 5,000 per side in epsg:3857)
 >""",
    default=None,
)
@click.option(
    "--year",
    type=int,
    help="Only extract datasets for a given year. This is a shortcut for --temporal-range=<int>--P1Y",
)
@click.option(
    "--temporal-range",
    type=str,
    help="Only extract datasets for a given time range, Example '2020-05--P1M' month of May 2020",
)
@click.option(
    "--frequency",
    type=str,
    help="Specify temporal binning: annual|annual-fy|semiannual|seasonal|nov-mar|apr-oct|all",
)
@click.option("--env", "-E", type=str, help="Datacube environment name")
@click.option(
    "-z",
    "complevel",
    type=int,
    default=6,
    help="Compression setting for zstandard 1-fast, 9+ good but slow",
)
@click.option(
    "--overwrite", is_flag=True, default=False, help="Overwrite output if it exists"
)
@click.option(
    "--tiles", help='Limit query to tiles example: "0:3,2:4"', callback=click_range2d
)
@click.option(
    "--debug",
    is_flag=True,
    default=False,
    hidden=True,
    help="Dump debug data to pickle",
)
@click.option(
    "--gqa",
    type=float,
    help="Only save datasets that pass `gqa_iterative_mean_xy <= gqa` test",
)
@click.option(
    "--dataset-filter",
    type=str,
    default=None,
    help='Filter to apply on datasets - {"collection_category": "T1"}'
)
@click.argument("products", type=str, nargs=1)
@click.argument("output", type=str, nargs=1, default="")
def save_tasks(
    grid,
    year,
    temporal_range,
    frequency,
    output,
    products,
    dataset_filter,
    env,
    complevel,
    overwrite=False,
    tiles=None,
    debug=False,
    gqa=None,
):
    """
    Prepare tasks for processing (query db).

    <todo more help goes here>

    \b
    Not yet implemented features:
      - output product config
      - multi-product inputs

    """
    from datacube import Datacube
    from .tasks import SaveTasks
    from .model import DateTimeRange

    filter = {}
    if dataset_filter:
        filter = json.loads(dataset_filter)

    if temporal_range is not None and year is not None:
        print("Can only supply one of --year or --temporal_range", file=sys.stderr)
        sys.exit(1)

    if temporal_range is not None:
        try:
            temporal_range = DateTimeRange(temporal_range)
        except ValueError:
            print(f"Failed to parse supplied temporal_range: '{temporal_range}'")
            sys.exit(1)

    if year is not None:
        temporal_range = DateTimeRange.year(year)

    if frequency is not None:
        if frequency not in ("annual", "annual-fy", "semiannual", "seasonal", "nov-mar", "apr-oct", "all"):
            print(f"Frequency must be one of annual|annual-fy|semiannual|seasonal|nov-mar|apr-oct|all and not '{frequency}'")
            sys.exit(1)

    dc = Datacube(env=env)
    products = products.split("+")
    if len(products) == 1:
        product = products[0]
        dss = None
        n_dss = None
    else:
        dss, n_dss, product, error_logger = _parse_products(dc, products, filter, temporal_range)

    if output == "":
        if temporal_range is not None:
            output = f"{product}_{temporal_range.short}.db"
        else:
            output = f"{product}_all.db"

    try:
        tasks = SaveTasks(
            output, grid, frequency=frequency, overwrite=overwrite, complevel=complevel
        )
    except ValueError as e:
        print(str(e))
        sys.exit(1)

    def on_message(msg):
        print(msg)

    def gqa_predicate(ds):
        return ds.metadata.gqa_iterative_mean_xy <= gqa

    predicate = None
    if gqa is not None:
        predicate = gqa_predicate

    try:
        ok = tasks.save(
            dc,
            product,
            dataset_filter=filter,
            temporal_range=temporal_range,
            tiles=tiles,
            predicate=predicate,
            debug=debug,
            msg=on_message,
            dss=dss,
            n_dss=n_dss,
        )
    except ValueError as e:
        print(str(e))
        sys.exit(2)

    if len(products) != 1:
        for product, count in error_logger.missing_counts.items():
            print(f"Product {product} has {count} missing datasets.")

    if not ok:
        # exit with error code, failure message was already printed
        sys.exit(3)


def _parse_products(dc, products, dataset_filter, temporal_range):

    query = dict(product=products, **dataset_filter)

    # TODO: find time range
    if temporal_range:
        query.update(temporal_range.dc_query(pad=0.6))
        dss = ordered_dss(dc, key=lambda ds: (ds.center_time, ds.metadata.region_code), **query)
        n_dss = min(dataset_count(dc.index, time=query["time"], product=product) for product in products)
    else:
        dss = dc.find_datasets(**query)
        dss.sort(key=lambda ds: (ds.center_time, ds.metadata.region_code))
        n_dss = min(dataset_count(dc.index, product=product) for product in products)

    paired_dss = groupby(dss, key=lambda ds: (ds.center_time, ds.metadata.region_code))
    error_logger = ErrorLogger(products)
    paired_dss = error_logger.filter(paired_dss)

    products = [dc.index.products.get_by_name(product) for product in products]
    fused_product = fuse_products(*products)
    map_fuse_func = lambda x: fuse_ds(*x, product=fused_product)
    dss = map(map_fuse_func, paired_dss)
    product = fused_product.name

    return dss, n_dss, product, error_logger


class ErrorLogger:

    def __init__(self, products):
        self.products = products
        self.missing_counts = dict((p, 0) for p in products)

    def append(self, ds_group):
        product_group = tuple(ds.type.name for ds in ds_group)
        for product in self.products:
            if product not in product_group:
                self.missing_counts[product] += 1

    def check(self, ds_group):
        return len(ds_group) == len(self.products)

    def filter(self, groups):
        for _, ds_group in groups:
            ds_group = tuple(ds_group)
            if not self.check(ds_group):
                self.append(ds_group)
            else:
                yield ds_group




import json

from odc.aio import S3Fetcher, s3_find_glob
from odc.stac import stac2ds
from pystac.item import Item


def bytes2stac(data):
    if isinstance(data, bytes):
        data = data.decode("utf-8")
    stac_doc = json.loads(data)
    return Item.from_dict(stac_doc)


def s3_fetch_dss(glob, s3=None):
    if s3 is None:
        s3 = S3Fetcher(aws_unsigned=True)

    blobs = s3(o.url for o in s3_find_glob(glob, skip_check=True, s3=s3))
    stacs = (bytes2stac(b.data) for b in blobs)

    return stac2ds(stacs)



"""
Fractional Cover Percentiles
"""
from functools import partial
from typing import Tuple
from itertools import product
import xarray as xr
import numpy as np
from odc.algo import keep_good_only
from odc.algo._percentile import xr_quantile_bands
from odc.algo._masking import _xr_fuse, _or_fuser, _fuse_mean_np, _fuse_or_np
from ._registry import StatsPluginInterface, register

NODATA = 255


class StatsFCP(StatsPluginInterface):
    
    NAME = "ga_fc_percentiles"
    SHORT_NAME = NAME
    VERSION = "0.0.1"
    PRODUCT_FAMILY = "fc_percentiles"

    def __init__(
        self,
        **kwargs
    ):
        super().__init__(input_bands=["water", "pv", "bs", "npv"], **kwargs)

    @property
    def measurements(self) -> Tuple[str, ...]:
        _measurments = [f"{b}_pc_{p}" for b, p in product(["pv", "bs", "npv"], ["10", "50", "90"])]
        _measurments.append("qa")
        return _measurments

    def native_transform(self, xx):
        """
        Loads data in its native projection. It performs the following:

        1. Load all fc and WOfS bands
        2. Set the high terrain slope flag to 0
        3. Set all pixels that are not clear and dry to NODATA
        4. Calculate the clear wet pixels
        5. Drop the WOfS band
        """
        
        water = xx.water & 0b1110_1111
        dry = water == 0
        xx = xx.drop_vars(["water"])
        xx = keep_good_only(xx, dry, nodata=NODATA)
        xx["wet"] = water == 128
        return xx

    def fuser(self, xx):
        wet = xx["wet"]
        xx = _xr_fuse(xx.drop_vars(["wet"]), partial(_fuse_mean_np, nodata=NODATA), '')

        band, *bands = xx.data_vars.keys()
        all_bands_invalid = xx[band] == NODATA
        for band in bands:
            all_bands_invalid &= xx[band] == NODATA

        xx["wet"] = _xr_fuse(wet, _fuse_or_np, wet.name) & all_bands_invalid
        return xx
    
    def reduce(self, xx: xr.Dataset) -> xr.Dataset:
        # (!all_bands_valid) & is_ever_wet => 0
        # (!all_bands_valid) & (!is_ever_wet) => 1
        # all_bands_valid => 2  

        wet = xx["wet"]
        xx = xx.drop_vars(["wet"])

        yy = xr_quantile_bands(xx, [0.1, 0.5, 0.9], nodata=NODATA)
        is_ever_wet = _or_fuser(wet).squeeze(wet.dims[0], drop=True)

        band, *bands = yy.data_vars.keys()
        all_bands_valid = yy[band] != NODATA
        for band in bands:
            all_bands_valid &= yy[band] != NODATA

        all_bands_valid = all_bands_valid.astype(np.uint8)
        is_ever_wet = is_ever_wet.astype(np.uint8)
        yy["qa"] = 1 + all_bands_valid - is_ever_wet * (1 - all_bands_valid)
        return yy


register("fc-percentiles", StatsFCP)
